/*
 * Non-blocking LED code inspired by
 * David A. Mellis (created 2005)
 * Paul Stoffregen (modified 8 Feb 2010)
 * Scott Fitzgerald (modified 11 Nov 2013)
 *
 * RM Learney November 2015
 *
 * 18 Jan 2016
 * Changed to use a state machine and with pin inputs suitable
 * for the LMP90080 + AFE4490 board Ver 3 Rev A
 *
 * Want to wake from sleep if the button has been pressed,
 * do something inside the button press state machine,
 * then go back to sleep.
 *
 * State transitions:
 *
 * OFF -> ON -> CALIBRATE -> ON -> OFF
 * or
 * OFF -> ON -> CALIBRATE -> OFF
 *
 * Single click -> OFF to ON or CALIBRATE to ON
 * Double click -> ON to CALIBRATE
 * Long press -> ON to OFF or CALIBRATE to OFF
 *
 * The initial OFF -> ON transition is automatic and within setup()
 *
 * March 2016
 * Added a startup routine to zero the BP sensor
 * Finalised RTD reading code and scaling - 100mV per degree C, accurate to 0.1C
 * Finalised BP code and scaling - 2000mV = 90mHg, 4000mV = 180mmHg
 */

#include <Arduino.h>
#include <SPI.h>
#include <Wire.h>
#include <EEPROM.h> // To read and write the on-chip EEPROM for BP calibration data
#include <avr/interrupt.h>
#include "AFE4490.h"
#include "LMP90080.h"
#include "DAC8564.h"
#include "MAX17043.h"
#include "LUT_3850.h" // Look-up table for a 3850ppm/K RTD
#include "boardfile.h"
#include "sleepFunctions.h"

#ifndef DEBUG //Some macros to aid debugging on a standard Arduino board
  #define _LEDHIGH PORTD |= 0b00010000; //SmartPipe V3 Rev A LED goes HIGH
  #define _LEDLOW  PORTD &= 0b11101111; //SmartPipe V3 Rev A LED goes LOW
  #define _LEDSET  (PIND & (1 << PD4))
#else
  #define _LEDHIGH PORTB |= 0b00100000; //Onboard Arduino LED goes HIGH
  #define _LEDLOW  PORTB &= 0b11011111; //Onboard Arduino LED goes LOW
  #define _LEDSET  (PINB & (1 << PB5))
#endif DEBUG

typedef enum { //SYSTEM_STATE
  WAKE,
  STARTUP,
  CALIBRATION,
  BATTERY_LEVEL,
  SLEEP
} deviceStates;

typedef enum { //BUTTON_STATE
  WAITING,
  PRESSED,
  RELEASED
} buttonPress;

typedef enum { //Rename these later
  BUTTON_BEGIN,
  STATE1,
  STATE2,
  STATE3,
  STATE4
} buttonStates; //BUTTON_RESPONSE

deviceStates systemState;
buttonPress button;
buttonStates buttonState;

volatile uint8_t oldPORTB;
volatile uint8_t oldPORTC;
volatile boolean newLMP = false; // DRDY from the LMP90080
volatile boolean newAFE = false; // DRDY from the AFE4490
volatile boolean lowBatt = false; // ALERT interrupt from the MAX17043
volatile boolean startedBlinking = false; //Have we waited long enough to start the battery indicator blinking?
volatile uint32_t _nonBlockingLEDtimer; // For blinking the LED in calibration or battery indicator mode

uint16_t _doubleClick = 160; //msec max timeout gap between double clicks
uint16_t _longPress = 1800; //msec which have to pass to detect a long press
uint16_t _calibrationFlash = 1000; //msec flashing for the calibration routine
#ifndef DEBUG
  uint16_t _batteryFlash = 200; //msec flashing for indicating the battery level
#else
  uint16_t _batteryFlash = 800; //extra delay to cope with the serial communications in DEBUG mode
#endif
uint8_t _flashNumber; //For displaying a number of flashes depending on battery charge state

int8_t accumulatorCount = 0; // The accumulator to capture the BP calibration
int32_t accumulator = 0;
volatile uint16_t _calibration; // For holding the rolling average value
volatile boolean newAccumulator = false;

volatile uint16_t _0mmHg = 0; // For holding 0mmHg offset
volatile float _1mmHg = 1.0; // For scaling the result coming from the BP sensor to match 2V/90mmHg (32000)

byte _lowBatt = 10; // Set the percentage at which the MAX17043 will shut down the system
const int8_t BP_EEPROM = 0; // The EEPROM address for the BP data MSB (LSB = this address +1)

uint16_t redValue; // For holding the red value of the AFE4490
uint16_t IRValue; // For holding the IR value of the AFE4490
int16_t sO2; //For holding the oxygen saturation value
int8_t battCharge; // For holding the battery charge level of the MAX17043

LMP90080 LMP = LMP90080();
AFE4490 AFE = AFE4490();
DAC8564 DAC = DAC8564();
MAX17043 batteryMonitor;
sleepFunctions Sleep;

void setup(){

  /* Turn the system on */
  pinMode(POWER, OUTPUT);   // Soft latch power line
  digitalWrite(POWER, HIGH); // Hold the POWER line high to keep the system on
  systemState = STARTUP; // But make sure we go straight to sleep after startup

  #ifdef DEBUG
    Serial.begin(115200);
  #endif

  /* Initialise SPI and I2C */
  SPI.begin();
  Wire.begin();

  /* Configure the Button indication LED */
  pinMode(LED, OUTPUT);
  startupLED();

  /* Initialise the Button */
  pinMode(SENSE, INPUT_PULLUP);
  button = WAITING; // Begin with the button waiting for input
  buttonState = BUTTON_BEGIN;

  /* Initialise the MAX17043 battery monitor and its ALERT line*/
  batteryMonitor.reset();
  batteryMonitor.quickStart();
  batteryMonitor.setAlertThreshold(_lowBatt); //Alert when only X% of charge remains

  /* Initialise the LMP90080 */
  LMP.init(CS_LMP);

  /* Initialise the DAC8564 */
  pinMode(LDAC, INPUT);
  AFE.triState();
  DAC.init(CS_DAC, 0); // DAC address 0
  DAC.write(DAC_ALL, 0x0000);

  AFE.triState();

  /* Import the BP calibration data from the EEPROM */
  //EEPROM.get(BP_EEPROM,_100mmHg);

  /* Initialise the AFE4490 */
  AFE.init(CS_AFE);

  /* Configure the Interrupts */
  cli(); // switch interrupts off while messing with their settings

  PCICR |= (1 << PCIE1);    // Enable PCINT1 ISR (Pin A1)
  PCMSK1 |= (1 << PCINT9);  //Enable PCINT9 CHANGE interrupt (Pin A1)
  oldPORTC = PINC; // Read the state of the button interrupt on startup

  PCICR |= (1 << PCIE0); //PB1 is in PCINT1 ISR (Pin PB1) - ALERT from MAX17043
  PCMSK0 |= (1 << PCINT1); //Enable PCINT1 CHANGE interrupt (Pin PB1)
  oldPORTB = PINB; // Read the state of the ALERT interrupt on startup

  sei();

  attachInterrupt(digitalPinToInterrupt(DRDY_AFE),AFE_Interrupt,RISING); //Note change in syntax
  attachInterrupt(digitalPinToInterrupt(DRDY_LMP),LMP_Interrupt,FALLING);

  Sleep.init();
}

void loop(){

  #ifdef DEBUG
    Serial.println("loop()");
  #endif

/* The system should spend most of its time in SLEEP */

  if(systemState == SLEEP) {
    #ifdef DEBUG
      Serial.println("Sleeping");
      delay(100);
    #endif
    _LEDLOW; //Make sure the LED doesn't stay on in sleep mode
    Sleep.lightSleep();
  }

/* Handle the push button */

  if(button != WAITING) { //The button handler decides when to go back to the WAITING state
    #ifdef DEBUG
      Serial.println("Button Handler");
    #endif
    buttonHandler();
  }

/* Handle the blinking LED for the Calibration Mode */

  if(systemState == CALIBRATION || systemState == STARTUP) {

    #ifdef DEBUG
      Serial.println("Calibration");
      Serial.println(_nonBlockingLEDtimer);
    #endif

    if( (millis() - _nonBlockingLEDtimer) >= _calibrationFlash) {
      nonBlockingLED();
    }
  }

  if(systemState == BATTERY_LEVEL) {
  /* Handle the blinking LED for Battery Indicator Mode
   *
   * 0  - 25%  = 1 flash
   * 26 - 50%  = 2 flashes
   * 51 - 75%  = 3 flashes
   * 76 - 100% = 4 flashes
   */
  #ifdef DEBUG
    Serial.println("Battery Level Indicator");
    Serial.print("Timer Duration: "); Serial.println(_nonBlockingLEDtimer);
    Serial.print("Flash Number: "); Serial.println(_flashNumber);
  #endif

  if(!startedBlinking) {

    if((millis() - _nonBlockingLEDtimer) >= 1200) { //Wait 1200msec after button release before starting the blinking

      startedBlinking = true;
    }
  }

  else {

    if(_flashNumber > 0) {

      if( (millis() - _nonBlockingLEDtimer) >= _batteryFlash) { // Should then flip straight into first 'flash'

        _flashNumber -= 1; // Need n*2 transitions to flash on and off
        nonBlockingLED();
      }
    }
    else {
      button = WAITING;
      systemState = SLEEP;
    }
  }
  }

/* Read sO2 Data from the AFE4490 */

  if(newAFE) {
    redValue = AFE.readLED(LED1_ALED1VAL);
    IRValue = AFE.readLED(LED2_ALED2VAL);

    AFE.triState();

    DAC.write(DAC_A,redValue);
    DAC.write(DAC_B,IRValue);

    AFE.triState();

    //sO2 = AFE.sO2calculation(redValue,IRValue);

    newAFE = false;
  }

/* Read BP and RTD Temperature Data from the LMP90080 */
/* SmartPipe Version 3 Rev A - BP on CH0 with VREF1, RTD on CH1 with VREF2 */

  if(newLMP) {

    #ifdef DEBUG
      Serial.println("LMP");
    #endif

    AFE.triState();

    int8_t Channel = LMP.getChan();

    if(Channel == 0) { // BP is on CH0 but inverted (VIN1-VIN0)

      uint16_t BP = LMP.read(LMP_DOUT);
      BP = ((0xFFFF - BP) - _0mmHg); // Because the results are inverted and offset

      if(BP > 0x8000) {
        BP = 0;
      }

      DAC.write(DAC_C, (uint16_t)(BP*_1mmHg));

      if(systemState == STARTUP) { // Get the 0 value for the BP sensor here at first startup

        if(newAccumulator == true) {
          _0mmHg = _calibration;
          _calibration = 0x0000; //Return _calibration to zero
          newAccumulator = false;
          systemState = SLEEP;
        }

        else { // Wait until the accumulator has collected some data
          rollingAverage(BP);
        }
      }

      if(systemState == CALIBRATION) {
        rollingAverage(BP);
      }
    }

    if(Channel == 1) { // RTD is on Channel 2 (VIN2-VIN3)

      uint16_t RTD = LMP.read(LMP_DOUT);
      DAC.write(DAC_D,LMP.getTemp(RTD)*160); //10mV/4096 * 65535 = approx 160 per degree, or 131 if 5V reference
    }

    AFE.triState();

    newLMP = false;
  }

/* Handle the ALERT pin from the MAX17043 */

  if(lowBatt) {
    batteryMonitor.clearAlert(); // Need to clear the ALERT bit in software
    powerDown();
  }

}


void buttonHandler(void) {
  /*
   * This function handles all states of the button press,
   * which changes behaviour of the main loop()
   *
   * On first entering the state machine, the button must have been pressed, hence CLICK
   * Following this, it determines whether there has been a single click,
   * double click or longer press.
   *
   * Want to wake from sleep if the button has been pressed,
   * do something inside the button press state machine,
   * then go back to sleep.
   *
   * State transitions:
   *
   * OFF -> ON -> CALIBRATE -> ON -> OFF
   *
   * Single click -> OFF to ON (by hardware)
   *              -> capture calibration in CALIBRATE mode
   *
   * Double click -> ON to CALIBRATE, CALIBRATE to ON
   *
   * Long press   -> ON to OFF, from any state including CALIBRATE - this can be changed
   *
   * The initial OFF -> ON transition is automatic and within setup()
   */

  uint32_t startTime;
  uint32_t wait;

  switch(buttonState) {

    case BUTTON_BEGIN: // Begin the state machine

      #ifdef DEBUG
        Serial.println("0");
      #endif

      if(button == PRESSED) {
        buttonState = STATE1;
      }

      else if(button == RELEASED) {
        systemState = SLEEP;
        buttonState = BUTTON_BEGIN; // Reset the button handling state machine
        button = WAITING; // Go back to waiting for another button press
        break;
        // Bit of a hack. The ISR sees the button release after a long press and jumps back in
        // to the button handler. This is the easiest way to deal with it for now.
      }

    case STATE1: // Waiting for button release

      #ifdef DEBUG
        Serial.println("1");
      #endif

      startTime = millis(); //Store the start time before the while loop

      while(button == PRESSED) {

        wait = millis() - startTime;

        if(wait >= _longPress) { // Long press functionality
          powerDown(); //Turn the system OFF
        }
      }

      buttonState = STATE2;

    //break;

    case STATE2: // Button just released. Is this a single or double click?

      #ifdef DEBUG
        Serial.println("2");
      #endif

      startTime = millis(); //Start waiting for the double click timeout now

      while(button == RELEASED) {

        wait = millis() - startTime;

        if(wait > _doubleClick) { // Single click functionality

          #ifdef DEBUG
            Serial.println("Click!");
          #endif

          if(systemState == CALIBRATION) {
            systemState = CALIBRATION;
            DAC.write(DAC_A,0xFFFF); //Visual display that the value has been captured
            //EEPROM.put(BP_EEPROM,_calibration); // Write calibration data to the EEPROM - takes 4msec
            _1mmHg = (32000.0/_calibration); // DAC code of 32000 = 2000mV at 90mmHg. Expects 180mmHg on pressure sensor
            buttonState = BUTTON_BEGIN; // Re-enter the state machine at the point of waiting for the button release
            button = WAITING; // Go back to waiting for another button press
            return; // Leave the button handler back to the main loop()
          }
          else if(systemState == BATTERY_LEVEL) { // If the button is clicked again within the battery level indicator
            systemState = SLEEP;
            buttonState = BUTTON_BEGIN; // Reset the button handling state machine
            button = WAITING; // Go back to waiting for another butto n press
            return;
          }
          else { //Single click switches into the battery level indicator state
            systemState = BATTERY_LEVEL;
            buttonState = BUTTON_BEGIN;
            button = WAITING;
            startedBlinking = false;

            power_twi_enable(); //Turn I2C circuitry on
            battCharge = batteryMonitor.getSoC();
            power_twi_disable(); //Turn I2C circuitry back off

            #ifdef DEBUG
              Serial.print("State of Charge:\t");Serial.println(battCharge);
            #endif

            if(battCharge > 0) { // 1% upwards +2 --%% this could be improved by starting at the ALERT level
              _flashNumber += 2;
            }
            if(battCharge > 25) { // 26% upwards + 2
              _flashNumber += 2;
            }
            if(battCharge > 50) { // 51% upwards + 2
              _flashNumber += 2;
            }
            if(battCharge > 75) { // 76% upwards + 2
              _flashNumber += 2;
            }

            return;
          }
        }
      }

      if(button == PRESSED) { //This would have to happen before the timeout
        buttonState = STATE3;
      }

    case STATE3: // Double Click - either enter or exit CALIBRATION mode

      #ifdef DEBUG
        Serial.println("3");
      #endif

      if(button == RELEASED) { // Double click functionality

        if(systemState == CALIBRATION) {
          systemState = SLEEP;
        }

        else {
          systemState = CALIBRATION;
          _nonBlockingLEDtimer = millis(); //Non-blocking LED blink
        }

        buttonState = BUTTON_BEGIN; // Reset the button handling state machine
        button = WAITING; //Go back to waiting for another button press - can only sleep again outside CALIBRATION mode
        return;
      }

    break;

    default:

      systemState = SLEEP;
      buttonState = BUTTON_BEGIN; // Reset the button handling state machine
      button = WAITING; //Go back to waiting for another button press

    break;

    }

  return;
}


void rollingAverage(uint16_t newData)
{
  #ifdef DEBUG
    Serial.println("Capture Calibration");
  #endif

  int8_t ignore; //Number of samples to ignore

  accumulatorCount++;

  if(systemState == STARTUP) {
    ignore = 2;
  }
  else if(systemState == CALIBRATION) {
    ignore = 1;
  }

  if(accumulatorCount >= ignore) {
    accumulator += newData;
  }

  if(accumulatorCount == (15 + ignore) ) { // Rolling 16-sample average (zero to 15), ignoring the first sample(s)
      accumulatorCount = 0; // Reset the counter
      _calibration = ((accumulator >> 4) & 0xFFFF); // 2^4 = 16
      accumulator = 0; //Reset the accumulator
      newAccumulator = true; //Just in case the sensor has burned out and the value reads 0x0000...
    }
}


void nonBlockingLED(void)
{
  #ifdef DEBUG
    Serial.println("Non-Blocking LED");
  #endif

  if(_LEDSET) {
    _LEDLOW
  }
  else {
    _LEDHIGH
  }

  _nonBlockingLEDtimer = millis();
}


void startupLED(void) //Can use blocking delay() because no timing issues in startup()
{
  delay(100);

  _LEDHIGH
  delay(20);
  _LEDLOW
  delay(60);
  _LEDHIGH
  delay(20);
  _LEDLOW
  delay(200);
  _LEDHIGH
  delay(20);
  _LEDLOW
  delay(60);
  _LEDHIGH
  delay(20);
  _LEDLOW


  #ifdef DEBUG
    Serial.println("startupLED");
    delay(100);
  #endif
}


void powerDown(void)
{
  #ifdef DEBUG
    Serial.println("Power Down");
  #endif

  PCICR = 0x00; // Disable all further pin change interrupts to lock-in the Power Down state

  for(int i=0;i<=16;i++) {
    _LEDLOW
    delay(60); //Can use blocking delay() because no timing issues during shutdown
    _LEDHIGH
    delay(10);
    _LEDLOW
  }

  digitalWrite(POWER,LOW);
}


void LMP_Interrupt(void)
{
  newLMP = true;
  Sleep.wakeupSPI();
}


void AFE_Interrupt(void)
{
  newAFE = true;
  Sleep.wakeupSPI();
}


ISR(PCINT0_vect) //For Pin PB1 - ALERT - which is PCINT1
{
    uint8_t changedbits;

    changedbits = PINB ^ oldPORTB;
    oldPORTB = PINB;

    if(changedbits & (1 << PINB1))
    {
        PCICR = 0x00; // Disable all further pin change interrupts to lock-in the Power Down state
        lowBatt = true;
        systemState = WAKE; // So it doesn't go straight back to SLEEP mode in the main loop()
    }
}


ISR(PCINT1_vect)
{
  /*
   * REMEMBER! The pin is default HIGH, and LOW when pressed, NOT the other way around
   * Pretty counterintuitive...
   *
   * Interestingly...the hardware design of the pin seems to automatically debounce it
   */

    uint8_t changedbits;

    changedbits = PINC ^ oldPORTC;

    if(changedbits & (1 << PINC1)) //PINC1 = Pin A1 was toggled
    {
      if(PINC & (1 << PINC1)) { //if the pin is now HIGH, it must have been LOW before
        button = RELEASED;
        _LEDLOW
      }
      else {
        button = PRESSED;
        _LEDHIGH

        if(systemState == SLEEP) {
          Sleep.wakeupTimer0();
          systemState = WAKE;
        }
      }
    }

    oldPORTC = PINC;
}
