/*
  AFE4490.cpp - Library for controlling TIs AFE4490 sO2 Analog Front-End
  Created by Robert M. Learney, October 10, 2013.

  Current version 15/09/15
  Now uses a constructor
  Better formatting and no 'extern' call
  Has a function to tri-state at the end of each read to allow multiple devices

  16/02/16
  Added int16_t sO2calculation(uint16_t I_660, uint16_t I_905) function
*/

#include "AFE4490.h"
#include "boardfile.h"
#include <math.h>

#define CONTROL2_DATA 0x020000 //0.5V TX_REF Voltage - best results for 3V supply (pages 76-77)

AFE4490::AFE4490(void)
{
}


boolean AFE4490::init(int8_t cs)
{ 
  _CS = cs;
  pinMode(_CS, OUTPUT);
  digitalWrite(_CS, HIGH);  //Start with CS held high
  _triState = CONTROL2_DATA;
  
  AFE4490::reset();

  AFE4490::write(CONTROL1, 0x000100); //bit 8 = Enable Timer Module, configure for (n+1) samples to be averaged
  AFE4490::write(CONTROL2, CONTROL2_DATA); 
  
  AFE4490::write(TIAGAIN, 0x000000); //LEDs same
  
  AFE4490::write(TIA_AMB_GAIN, 0x00C200); //1kHz filter corner, IR LED Stage 2 ON, Stage 2 Gain = x3, TIA 250kOhm
  
  AFE4490::write(LEDCNTRL,0x010302);  //Sets '01' in LED_RANGE(pg 76) for 0.5V TX_REF (pg 77), then both LED currents
  
  AFE4490::setTimers();
  //AFE4490::Oct14Timers();
 
  return true;
}


boolean AFE4490::reset(void)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));
  
  digitalWrite(_CS, LOW);  
  
  SPI.transfer(CONTROL0);    //Software reset for the AFE4490 in CONTROL0 (0x00)
  SPI.transfer(0x00);
  SPI.transfer(0x00);
  SPI.transfer(0x08);
  
  digitalWrite(_CS, HIGH);

  return true;
}


uint32_t AFE4490::read(uint8_t thisRegister)
{ 
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));
  
/*
  Read the data back from a normal register in the AFE4490. Make sure the READ bit in the CONTROL register is set!!!
  This is the sequence to be followed: 
  1. First set the READ bit in the CONTROL0 register
  2. Send the 8-bit address to be read back
  3. The next 24 bits on MISO will be the contents of the register in question
  4. Turn off the READ bit in the CONTROL0 register
*/ 

  uint32_t retVal;

  digitalWrite(_CS, LOW);     
  
  SPI.transfer(CONTROL0);                    //This sequence sets the READ bit in CONTROL0 (0x00)  
  SPI.transfer(0x00);
  SPI.transfer(0x00);  
  SPI.transfer(0x01);                   
 
  SPI.transfer(thisRegister);                //Send the register to be read back
  retVal |= (SPI.transfer(0x00) << 16);
  retVal |= (SPI.transfer(0x00) << 8);
  retVal |= SPI.transfer(0x00);

  SPI.transfer(CONTROL0);                    //This sequence clears the READ bit in CONTROL0 (0x00)
  SPI.transfer(0x00);
  SPI.transfer(0x00);
  SPI.transfer(0x00);
  
  digitalWrite(_CS, HIGH);
  
  return retVal;
}


uint16_t AFE4490::readLED(uint8_t thisRegister)
{ 
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));
  
/*
  Read the data back from an LED register in the AFE4490. Make sure the READ bit in the CONTROL register is set!!!
  This is the sequence to be followed: 
  1. First set the READ bit in the CONTROL0 register
  2. Send the 8-bit address to be read back
  3. The next 24 bits on MISO will be the contents of the register in question
  4. Turn off the READ bit in the CONTROL0 register
*/ 
  uint8_t rxBuffer[5] = {0};
  uint32_t temp = 0;
  
  digitalWrite(_CS, LOW);     
  
  SPI.transfer(CONTROL0);                    //This sequence sets the READ bit in CONTROL0 (0x00)  
  SPI.transfer(0x00);
  SPI.transfer(0x00);  
  SPI.transfer(0x01);                   

  SPI.transfer(thisRegister);
  rxBuffer[1] = SPI.transfer(0x00);
  rxBuffer[2] = SPI.transfer(0x00);
  rxBuffer[3] = SPI.transfer(0x00);

  SPI.transfer(CONTROL0);                    //This sequence clears the READ bit in CONTROL0 (0x00)
  SPI.transfer(0x00);
  SPI.transfer(0x00);
  SPI.transfer(0x00);
  
  digitalWrite(_CS, HIGH);

  /*
   * Testing alternative logic to construct the returned value. 
   * According to: http://www.avrfreaks.net/forum/twos-compliment-straight-binary-stuck
   * In order to shift a 16-bit 2s complement value (-32,768 to +32,767) into straight binary, need to add
   * 0x8000 to the result (shifting up by 32,768 so the range is 0-65,536)
   * 
   * So with a 22-bit value, need to add 0x200000 and then shift down by 5 bits and then take the lower 16 bits
   * 
   * Shifting by 5 bits instead of 6 makes sure we use the full 0-5V range instead of 2.5-5V (the first bit will always* be 1)
   * 
   * *almost always
   */

  temp = rxBuffer[1];
  temp = (temp << 8) | rxBuffer[2];
  temp = (temp << 8) | rxBuffer[3];
  temp += 0x200000;
  return ( (temp >> 5) & 0xFFFF);
}


boolean AFE4490::write(uint8_t thisRegister, uint32_t thisValue)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));
  
  digitalWrite(_CS, LOW);
  
  SPI.transfer(thisRegister);
  SPI.transfer(((thisValue >> 16) & 0xFF));
  SPI.transfer(((thisValue >> 8)  & 0xFF));
  SPI.transfer((thisValue & 0xFF));
  
  digitalWrite(_CS, HIGH);

  return true;
}


boolean AFE4490::triState(void)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));
  //Toggle the tristate bit in CONTROL2

  _triState = _triState^0x000400;                  //Toggle only the 10th bit = SPI tristate bit

  digitalWrite(_CS, LOW);
  
  SPI.transfer(CONTROL2);
  SPI.transfer(((_triState >> 16) & 0xFF));
  SPI.transfer(((_triState >> 8)  & 0xFF));
  SPI.transfer((_triState & 0xFF));
  
  digitalWrite(_CS, HIGH);

  return true;
}


int16_t AFE4490::sO2calculation(uint16_t I_660, uint16_t I_905) {

  /*
   * Molar Extinction Coefficients of Hb and HbO2 (cm^-1 mM^1) taken from:
   * Matcher SJ (1995) Analytical Biochemistry 227(1);54-68
   * via Dept of Medical Physics at UCL - hemoglobin in cuvette
   * http://www.ucl.ac.uk/medphys/research/borl/intro/spectra
   * 
   * e_Hb_660   = 3.4408
   * e_Hb_905   = 0.8944
   * e_HbO2_660 = 0.3346
   * e_HbO2_905 = 1.3452
   * 
   * Molar Extinction Coefficients (cm^-1 M^-1) from S.Prahl,
   * University of Waterloo
   * http://www.npsg.uwaterloo.ca/data/blood.php
   * 
   * e_Hb_660   = 3227.0
   * e_Hb_905   = 768.59
   * e_HbO2_660 = 319.6
   * e_HbO2_905 = 1207.5
   * 
   * NB 
   * specific extinction coefficient = base 10 log units
   * specific absorption coefficient = natural log units
   * Therefore scaling factor = ln(10) = 2.3025851
   * absorption coefficient = ln(10) * extinction coefficient
   * 
   * To convert *molar* extinction coefficients to absorption coefficients:
   * Assume tested with 15g/dL Hb
   * 
   * mu_a(lambda) = ln(10) * molar ext coeff(lambda) * (150 g Hb/L)/(64,458 g Hb/mole)
   * 
   * mu_a(lambda) = 0.005358338204 * molar extinction coefficient
   * 
   * But don't forget [Total Hb] affects scattering...and these probably aren't 15g/L measurements
   * 
   * Using University of Waterloo Values (more recent than Matcher et al)
   * e_Hb_660   = 3227.0 * ln(10)* (150/64458) = 17.291
   * e_Hb_905   = 768.59 -> 4.118
   * e_HbO2_660 = 319.6 -> 1.713
   * e_HbO2_905 = 1207.5 -> 6.470
   * 
   * The values for plasma were originally given as effective absorbtion coefficients in um^-1 
   * From Meinke, Muller, Friebel and Helfmann J. Biomed. Opt., 12(1):014024-1-014024-9, 2007.
   * Most absorbance in plasma is apparently from bilirubin pigments
   * 
   * The values for H2O were given as specific absorption coefficients in um^-1
   * K.F. Palmer and D. Williams. Optical properties of water in the near infrared. 
   * J. Opt. Soc. Am., 64:1107-1110, August 1974.
   * 
   * I then multiplied each by 10,000 to get back to coefficients in cm^-1, same as 
   * the HbO2 and Hb values
   */
   
  float e_Hb_660      = 17.291;
  float e_Hb_905      = 4.118;
  float e_HbO2_660    = 1.713;
  float e_HbO2_905    = 6.470;
  float e_H2O_660     = 4.1E-3;
  float e_H2O_905     = 7.147E-2;
  float e_plasma_660  = 1.515E-2;
  float e_plasma_905  = 8.56E-3; //Why is plasma absorbance at 905 lower than that for H2O?

  float haematocrit = 0.45; // Assume 45% Hct if Hb 15g/dL

  uint16_t I0_660 = 1111; //Placeholder
  uint16_t I0_905 = 1111; //Placeholder

  float A_660 = -1*log10(I_660/I0_660); //Absorbance at 660nm
  float A_905 = -1*log10(I_905/I0_905); //Absorbance at 905nm

  /*
   * Now remember, the absorbance at any given wavelength:
   * 
   * A_660 = e_HbO2_660[HbO2] + e_Hb_660[Hb] + e_plasma_660[Plasma] + scatter @ 660
   * A_905 = e_HbO2_905[HbO2] + e_Hb_906[Hb] + e_plasma_905[Plasma] + scatter @ 905
   * 
   * NB assume the path length is equal for all wavelengths
   * 
   * [HbO2] + [Hb] = [Total Hb]
   * [Total Hb] + [Plasma] = a fixed value
   * 
   * In order to solve for [Plasma] as well as [HbO2] and [Hb],
   * I need an extra LED wavelength to get the haematocrit...
   * 
   * Or can I assume a haematocrit of 40% and ignore scatter:
   * 
   * A_660 = e_660*[HbO2] + e_660*[Hb] + e_plasma_660*(0.6)
   * A_905 = e_905*[HbO2] + e_905*[Hb] + e_plasma_905*(0.6)
   * 
   */

  A_660 = A_660 - ( (1.0 - haematocrit) * e_H2O_660); // Use H2O values but could change to plasma
  A_905 = A_905 - ( (1.0 - haematocrit) * e_H2O_905);   
  
  /* 
   * Solve as a matrix...
   * 
   * |A_660| = |e_HbO2_660  e_Hb_660| *  |HbO2|
   * |A_905|   |e_HbO2_905  e_Hb_905|    | Hb |
   * 
   * So:
   * 
   * |HbO2| = |e_HbO2_660  e_Hb_660|^-1 * |A_660|
   * | Hb |   |e_HbO2_905  e_Hb_905|      |A_905|
   * 
   * Inverse of a 2x2 matrix is:
   * 
   * |a b| = (1/det) * |d  -b|  where determinant is (ad - bc)
   * |c d|             |-c  a|
   * 
   */

  float a = e_HbO2_660;
  float b = e_Hb_660;
  float c = e_HbO2_905;
  float d = e_Hb_905;
  
  float det = ( (a*d) - (b*c) );
   
  float HbO2 = ( (d/det) * A_660 ) - ( (b/det) * A_905 );
  float Hb   = ( (a/det) * A_905 ) - ( (c/det) * A_660 ); // (-c * A_660) and (a * A_905)

  return (uint16_t)(100 * HbO2/(HbO2 + Hb)); // Oxygen saturation
}


boolean AFE4490::setTimers(void)
{  
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));
  
/*
 * Allow the user to pass in a pulse frequency as a fixed number of options:
 * NB the numbers represent ticks of the AFE4490's internal 4MHz clock
 * _62Hz = 62.5Hz = 64000
 * _125Hz = 32000
 * _250Hz = 16000
 * _500Hz = 8000
 * _750Hz = 4000
 * _1000Hz = 2000
 * 
 * These will pre-select the PRP register's value (register 29), from which all the others are derived
 */

 /* 
  * This version (20/11/15) tries to cut power expenditure by 'front loading' all LED & ADC activity to the beginning
  * of the PRP
  * 
  * The base rate (PRPcount) is that for a 500Hz cycle (8000) i.e. each LED only flashes for 500usec
  * 
  * The rest of the period between the end of the 2nd LED pulse & conversion until the end of the PRP is then just
  * 'dead air', where nothing happens from count 7999 to 63999.
  * 
  * This still means each 'block' of LED activity repeats at 62.5Hz (64000 counts of 4MHz clock)
  */
  
  AFE4490::write(ADCRSTSTCT0,     0);
  AFE4490::write(ADCRSTENDCT0,    6);
  AFE4490::write(LED2CONVST,      8); //Sart ADC for IR signal
  AFE4490::write(ALED2STC,        50); //Start sampling ambient IR (1600 ticks)
  AFE4490::write(ALED2ENDC,     1650); //Stop sampling ambient IR
  AFE4490::write(LED2CONVEND,    508); //Stop ADC for IR signal (500 ticks)

  AFE4490::write(ADCRSTSTCT1,   16000);
  AFE4490::write(LED1LEDSTC,    16000); //Turn on Red LED
  AFE4490::write(ADCRSTENDCT1,  16006);
  AFE4490::write(ALED2CONVST,   16008); //Start ADC for ambient IR
  AFE4490::write(LED1STC,       16050); //Start sampling Red LED (1600 ticks)
  AFE4490::write(LED1ENDC,      17650); //Stop sampling Red LED
  AFE4490::write(LED1LEDENDC,   17652); //Turn off Red LED (1652 ticks)
  AFE4490::write(ALED2CONVEND,  16508); //Stop ADC for ambient IR (500 ticks)

  AFE4490::write(ADCRSTSTCT2,   32000); 
  AFE4490::write(ADCRSTENDCT2,  32006);
  AFE4490::write(LED1CONVST,    32008); //Start ADC for Red LED
  AFE4490::write(ALED1STC,      32050); //Start sampling ambient Red (1600 ticks)
  AFE4490::write(ALED1ENDC,     33650); //Finish sampling ambient Red
  AFE4490::write(LED1CONVEND,   32508); //Stop ADC for Red LED (500 ticks)

  AFE4490::write(ADCRSTSTCT3,   48000);
  AFE4490::write(LED2LEDSTC,    48000); //Turn on IR LED
  AFE4490::write(ADCRSTENDCT3,  48006);
  AFE4490::write(ALED1CONVST,   48008); //Start ADC for ambient Red
  AFE4490::write(LED2STC,       48050); //Start sampling IR (1600 ticks)
  AFE4490::write(LED2ENDC,      49650); //Finish sampling IR
  AFE4490::write(LED2LEDENDC,   49652); //Turn off IR LED (1652 ticks)
  AFE4490::write(ALED1CONVEND,  48508); //End ADC for ambient Red (500 ticks)

  AFE4490::write(PRPCOUNT,      63999);

  return true;
}

//
//boolean AFE4490::setTimers(void)
//{  
//  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));
//  
///*
// * Allow the user to pass in a pulse frequency as a fixed number of options:
// * NB the numbers represent ticks of the AFE4490's internal 4MHz clock
// * _62Hz = 62.5Hz = 64000
// * _125Hz = 32000
// * _250Hz = 16000
// * _500Hz = 8000
// * _750Hz = 4000
// * _1000Hz = 2000
// * 
// * These will pre-select the PRP register's value (register 29), from which all the others are derived
// */
//
// /* 
//  * This version (20/11/15) tries to cut power expenditure by 'front loading' all LED & ADC activity to the beginning
//  * of the PRP
//  * 
//  * The base rate (PRPcount) is that for a 500Hz cycle (8000) i.e. each LED only flashes for 500usec
//  * 
//  * The rest of the period between the end of the 2nd LED pulse & conversion until the end of the PRP is then just
//  * 'dead air', where nothing happens from count 7999 to 63999.
//  * 
//  * This still means each 'block' of LED activity repeats at 62.5Hz (64000 counts of 4MHz clock)
//  */
// 
// 
//  uint16_t PRPcount = 8000;
//
//  uint16_t Period1 = PRPcount/4;
//  uint16_t Period2 = PRPcount/2;
//  uint16_t Period3 = 3*Period1;
//  uint16_t Period4 = PRPcount;
//  uint16_t endPRP = 64000;
//  
//  AFE4490::write(1,(Period3 +50));
//  AFE4490::write(2,(Period4 -2));
//  AFE4490::write(3,Period3);
//  AFE4490::write(4,(Period4 -1));
//  AFE4490::write(5,50);
//  AFE4490::write(6,(Period1 -2));
//  AFE4490::write(7,(Period1 +50));
//  AFE4490::write(8,(Period2 -2)); 
//  AFE4490::write(9,Period1); 
//  AFE4490::write(10,(Period2 -1)); 
//  AFE4490::write(11,(Period2 +50)); 
//  AFE4490::write(12,(Period3 -2)); 
//  AFE4490::write(13,4);
//  AFE4490::write(14,(Period1 -1));
//  AFE4490::write(15,(Period1 +4));
//  AFE4490::write(16,(Period2 -1));
//  AFE4490::write(17,(Period2 +4));
//  AFE4490::write(18,(Period3 -1));
//  AFE4490::write(19,(Period3 +4));
//  AFE4490::write(20,(Period4 -1));
//  AFE4490::write(21,0);
//  AFE4490::write(22,6);
//  AFE4490::write(23,Period1); 
//  AFE4490::write(24,(Period1 +3)); 
//  AFE4490::write(25,Period2); 
//  AFE4490::write(26,(Period2 +3)); 
//  AFE4490::write(27,Period3); 
//  AFE4490::write(28,(Period3 +3)); 
//  AFE4490::write(29,(endPRP -1));
//
//  return true;
//}

boolean AFE4490::Oct14Timers(void)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));

  //Values taken from page 36 of the October 2014 datasheet
  AFE4490::write(1,6050);
  AFE4490::write(2,7998);
  AFE4490::write(3,6000);
  AFE4490::write(4,7999);
  AFE4490::write(5,50);
  AFE4490::write(6,1998);
  AFE4490::write(7,2050);
  AFE4490::write(8,3998);
  AFE4490::write(9,2000);
  AFE4490::write(10,3999);
  AFE4490::write(11,4050);
  AFE4490::write(12,5998);
  AFE4490::write(13,4);
  AFE4490::write(14,1999);
  AFE4490::write(15,2004);
  AFE4490::write(16,3999);
  AFE4490::write(17,4004);
  AFE4490::write(18,5999);
  AFE4490::write(19,6004);
  AFE4490::write(20,7999);
  AFE4490::write(21,0);
  AFE4490::write(22,3);
  AFE4490::write(23,2000);
  AFE4490::write(24,2003);
  AFE4490::write(25,4000);
  AFE4490::write(26,4003);
  AFE4490::write(27,6000);
  AFE4490::write(28,6003);
  AFE4490::write(29,7999);

  return true;
}
