/*
 LMP90080.cpp - Header file for controlling TI's LMP90080 multi-channel AFE
 
 Created by Robert M. Learney, September 1, 2015.

 Changelogs
 15/9/15
 Added the lowpass filter to the temperature calculation routine

 SmartPipe Version 3 Revision A

 BP is connected to CH0 (VIN0 - VIN1), using VREF1
 RTD is connected to CH1 (VIN2 - VIN3), using VREF2
 */

#include "LMP90080.h"
#include "boardfile.h"

LMP90080::LMP90080(void)
{
}


void LMP90080::init(int8_t cs)
{ 
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0));

  _CS = cs;
  pinMode(_CS, OUTPUT);
  digitalWrite(_CS, HIGH);  //Start with CS held high
  
  LMP90080::resetCalibrations();
  delay(10);

  LMP90080::write(SPI_HANDSHAKECN, 0b00001001); //SDO isn't used as DRDY, SDO goes high-Z after the last falling edge of SCLK
  LMP90080::write(SPI_DRDYBCN, 0b10000011); //SPI_DRDYBCN = D6 as DRDY, Correct FGA gain error with last known coefficients

  LMP90080::write(BGCALCN, 0x03); //Continuous background gain and offset estimation
  LMP90080::write(ADC_AUXCN, 0x05); //Set current sources to 500uA
  
  LMP90080::channelInput(CH0, 0b00001000);//BP is on Channel 0 reversed (VIN1-VIN0), VREF1, Burnout OFF
  LMP90080::channelConfig(CH0, (_13SPS | FGA64 | BUFFER_ON)); //BP is on Channel 0 reversed (VIN1-VIN0)
  LMP90080::channelInput(CH1, 0b01010011); //RTD is on Channel 1 (VIN2-VIN3), VREF2, Burnout OFF  
  LMP90080::channelConfig(CH1, (_13SPS | FGA8 | BUFFER_ON)); //RTD is on Channel 1 (VIN2-VIN3) 

  LMP90080::offsetCorrection();

  LMP90080::channelConfig(CH0, (_214SPS | FGA64 | BUFFER_ON)); //BP back to normal sampling rate
  LMP90080::channelConfig(CH1, (_26SPS | FGA8 | BUFFER_ON)); //RTD back to normal sampling rate 
  LMP90080::write(CH_SCAN, (SCANMODE2 | (CH1 << 3) | CH0) ); //Scanmode 2, from CH0 to CH1 
}


void LMP90080::offsetCorrection(void)
{ 
  /*
   * For some reason if the sampling rate for the BP is set to 26SPS the system goes strange 
   * and the offset is 1/2 what it should be, leading to some weird results...
   */
   
  _BPoffset = LMP90080::read(offsetCH0);
  _RTDoffset = LMP90080::read(offsetCH1);  

  LMP90080::write(CH_SCAN, (SCANMODE0 | (CH0 << 3) | CH0) ); //Just read channel 0
  
  LMP90080::write(SCALCN, 0x1); //Offset coefficient determination
  delay(100); //Wait until the calibration is complete
}


void LMP90080::BPgainCorrection(void)
{
  //Expects the pressure to be at 100mmHg
  _RTDgain = LMP90080::read(gainCH1);
  
  LMP90080::channelConfig(CH0, (_13SPS | FGA64 | BUFFER_ON)); //Slow down the BP read
  LMP90080::write(CH_SCAN, (SCANMODE0 | (CH0 << 3) | CH0) ); //Just read channel 0
  LMP90080::write16(gainCH0, 20000); //So 100mmHg = code of 20000
  LMP90080::write(SCALCN,0x2); //Put the system into Gain Coefficient Determination mode
  while(LMP90080::readByte(SCALCN)) {
    //Do nothing until the SCALCN bit is cleared to 0, indicating calibration is over
  }
  LMP90080::channelConfig(CH0, (_214SPS | FGA64 | BUFFER_ON)); //Return to normal BP sampling rate
  LMP90080::write(CH_SCAN, (SCANMODE2 | (CH1 << 3) | CH0) ); //Scanmode 2, from CH0 to CH1 
  //LMP90080::write16(gainCH1,_RTDgain); //Just in case this was overwritten...
}

void LMP90080::reset(void)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 

  LMP90080::write(0x00, 0xC3);
}


void LMP90080::resetCalibrations(void)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 

  LMP90080::write(ADC_AUXCN, 0b01000000); //Allow the reset() function to reset the calibration registers
  LMP90080::write(0x00, 0xC3);
}


void LMP90080::powerOn(void)
{
  LMP90080::write(0x08, 0x00);
}


void LMP90080::powerOff(void) //Send the LMP90080 into power-down mode
{
  LMP90080::write(0x08, 0x10);
}


void LMP90080::channelInput(uint8_t channel, uint8_t data)
{
  uint8_t _CH = 0x00;

  switch (channel) {
    case 0:
      _CH = 0x20;
      break;
    case 1:
      _CH = 0x22;
      break;
    case 2:
      _CH = 0x24;
      break;
    case 3:
      _CH = 0x26;
      break;
    case 4:
      _CH = 0x28;
      break;
    case 5:
      _CH = 0x2A;
      break;
    case 6:
      _CH = 0x2C;
      break;
    default:
      return;
      break;
  }

    LMP90080::write(_CH, data);
}


void LMP90080::channelConfig(uint8_t channel, uint8_t data)
{
  uint8_t _CH = 0x00;

  switch (channel) {
    case 0:
      _CH = 0x21;
      break;
    case 1:
      _CH = 0x23;
      break;
    case 2:
      _CH = 0x25;
      break;
    case 3:
      _CH = 0x27;
      break;
    case 4:
      _CH = 0x29;
      break;
    case 5:
      _CH = 0x2B;
      break;
    case 6:
      _CH = 0x2D;
      break;
    default:
      return;
      break;
  }

    LMP90080::write(_CH, data);
}


int8_t LMP90080::readByte(uint8_t address)
{
  int8_t data;
  
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  SPI.transfer(0x10); //Write the address to be read (NB!)
  SPI.transfer(0x00 | (address >> 4)); //Send URA to be read
  SPI.transfer(0x80 | (address & 0x0F)); //Command to read 1 byte (0b1000) then the LRA
  data = SPI.transfer(0x00);
  digitalWrite(_CS,HIGH);
  
  return data;
}


uint16_t LMP90080::read(uint8_t address)
{ 
  uint16_t data;
  
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  SPI.transfer(0x10); //Write the address to be read (NB!)
  SPI.transfer(0x00 | (address >> 4)); //Send URA to be read
  SPI.transfer(0xA0 | (address & 0x0F)); //Command to read 2 bytes (0b1010) then the LRA
  data = (SPI.transfer(0x00) << 8);
  data |= SPI.transfer(0x00);
  digitalWrite(_CS,HIGH);
  
  return data;
}


boolean LMP90080::write(uint8_t address, uint8_t data0)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 

  digitalWrite(_CS,LOW);
  SPI.transfer(0x10); //Write URA command
  SPI.transfer(0x00 | (address >> 4)); //Send the URA
  SPI.transfer(0x00 | (address & 0x0F)); //Send the LRA
  SPI.transfer(data0);
  digitalWrite(_CS,HIGH);

  return true;
}


boolean LMP90080::write16(uint8_t address, uint16_t data)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 

  digitalWrite(_CS,LOW);
  SPI.transfer(0x10); //Write URA command
  SPI.transfer(0x00 | (address >> 4)); //Send the URA
  SPI.transfer(0x20 | (address & 0x0F)); //Send the LRA + 'write 2 bytes' command
  SPI.transfer( ((data >> 8) & 0xFF) ); //Send the upper byte
  SPI.transfer( (data & 0xFF) ); //Send the lower byte
  digitalWrite(_CS,HIGH);

  return true;
}

/*
boolean LMP90080::write(uint8_t address, uint8_t data0, uint8_t data1, uint8_t data2)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 

  digitalWrite(_CS,LOW);
  SPI.transfer(0x10); //Write URA command
  SPI.transfer(0x00 | (address >> 4)); //Send the URA
  SPI.transfer(0x40 | (address & 0x0F)); //Send the LRA + 'write 3 bytes' command
  SPI.transfer(data0);
  SPI.transfer(data1);
  SPI.transfer(data2);
  digitalWrite(_CS,HIGH);

  return true;
}
*/

int8_t LMP90080::getChan(void) //Get the channel number whose conversion triggered the DRDY
{
//Channel number of the latest conversion is held in the last 3 bits of register 0x19 (SENDIAG_FLAGS)

  int8_t channel = 0x00;

  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  SPI.transfer(0x10); //Write the address to be read (NB!)
  SPI.transfer(0x01); //Send URA to be read
  SPI.transfer(0x89); //Send the LRA and the 'read' code
  channel = (SPI.transfer(0x00) & 0x07); //0x07 = 0b00000111 - mask out the bottom 3 bits
  digitalWrite(_CS,HIGH);

  return channel;  
}


uint16_t LMP90080::calculateTemp(int16_t RTDcode)
{
  // It might be easier/better to use a look-up table here than live calculations
  const float A  = 3.9083E-3; //These coefficients are for a 3850ppm/K SMD RTD
  const float B  = -5.775E-7;
  const float R0 = 100.0; //Value of the RTD at 0 celsius
  float R_RTD;
  float celsius;

  /* With the LMP90080, the 4-wire RTD value is calculated according to Application Note 2265 from TI
   * This is different from the methodology for the ADS1148/ADS1248 series
   * 
   * Output Code  = (RTD * Gain * 2^15-1) / Reference Resistor Value
   * 
   * So RTD = (Output Code * 1000) / (Gain * 2^15)
   * 
   * So pre-calculate 1000/(Gain*32768) 
   * = 0.01525878906 or 1.525587E-2 with gain = 2
   * = 0.007629394531 or 7.629394E-3 with gain = 4
   * = 0.003814697266 or 3.814697E-3 with gain = 8
   */
  
  R_RTD = (float)RTDcode * 3.81469E-3; //Truncated to 6 digits total - results seem more accurate
  celsius = (-1*A) + ( sqrt( (A*A) + (4*B*((R_RTD/R0)-1) ) ) ); //+ 4ac because c is negative - don't forget the signs!
  celsius = celsius/(2.0*B);

  /* http://www-users.cs.york.ac.uk/~fisher/mkfilter
   * Butterworth lowpass filter
   * 2nd order
   * sample rate = 13.42 Hz - using internal clock on LMP90080
   * corner frequency = 0.5 Hz
   * Note all floating point values truncated to 6 digits total
   */

  float dCoefficient1 = -0.71818;
  float dCoefficient2 = 1.67131;
  float dGain = 8.53516E+01;

  xvCelsius[0] = xvCelsius[1]; 
  xvCelsius[1] = xvCelsius[2]; 
  xvCelsius[2] = celsius / dGain;
  yvCelsius[0] = yvCelsius[1];
  yvCelsius[1] = yvCelsius[2]; 
  yvCelsius[2] = (xvCelsius[0] + xvCelsius[2]) + 2*xvCelsius[1]
                 + (dCoefficient1 * yvCelsius[0]) 
                 + (dCoefficient2 * yvCelsius[1]);
                 
  return (uint16_t)(100.0*yvCelsius[2]); //Scale by 100 and turn back into a uint16_t
}


uint16_t LMP90080::filterBessel_100Hz_5(int16_t BPcode)
{
   /* http://www-users.cs.york.ac.uk/~fisher/mkfilter
   * Bessel lowpass filter - no ringing or overshoot compared to butterworth, linear phase shift
   * 5th order
   * sample rate = 214.64 Hz - using internal clock on LMP90080
   * corner frequency = 100 Hz
   * Note all floating point values truncated to 6 digits total
   * 
   * Experiment showed this to be less optimal than the 30Hz 3-pole Butterworth for BP
   */
  
  float dGain = 1.29342;
  float dCoefficient1 = -0.59325;
  float dCoefficient2 = -3.28497;
  float dCoefficient3 = -7.28409;
  float dCoefficient4 = -8.08538;
  float dCoefficient5 = -4.49299;

  xvBP[0] = xvBP[1]; xvBP[1] = xvBP[2]; xvBP[2] = xvBP[3]; xvBP[3] = xvBP[4]; xvBP[4] = xvBP[5]; 
  xvBP[5] = (float)BPcode / dGain;
  yvBP[0] = yvBP[1]; yvBP[1] = yvBP[2]; yvBP[2] = yvBP[3]; yvBP[3] = yvBP[4]; yvBP[4] = yvBP[5]; 
  yvBP[5] = (xvBP[0] + xvBP[5]) + 5*(xvBP[1] + xvBP[4]) + 10*(xvBP[2] + xvBP[3])
                     + (dCoefficient1 * yvBP[0]) + (dCoefficient2  * yvBP[1])
                     + (dCoefficient3 * yvBP[2]) + (dCoefficient4 * yvBP[3])
                     + (dCoefficient5 * yvBP[4]);

  uint16_t retVal = (uint16_t)yvBP[5];
  
  if(retVal > 0x7FFF) //Clip those crazy negative values (or 65535 etc) to zero 
  {
    retVal = 0x0000;
  }
  
  return retVal;
}


uint16_t LMP90080::filterButter_30Hz_3(int16_t code)
{
   /* http://www-users.cs.york.ac.uk/~fisher/mkfilter
   * Butterworth lowpass filter - rolls off faster than Bessel
   * 3rd order
   * sample rate = 214.64 Hz - using internal clock on LMP90080
   * corner frequency = 30 Hz
   * Note all floating point values truncated to 6 digits total
   * 
   * Experiment showed this to be an optimal combination of smoothness
   * and filter complexity for BP results
   */
  
  float dGain = 23.9761;
  float dCoefficient1 = 0.160307;
  float dCoefficient2 = -0.777034;
  float dCoefficient3 = 1.28306;

  xvBP[0] = xvBP[1]; xvBP[1] = xvBP[2]; xvBP[2] = xvBP[3];
  xvBP[3] = (float)code / dGain;
  yvBP[0] = yvBP[1]; yvBP[1] = yvBP[2]; yvBP[2] = yvBP[3];; 
  yvBP[3] = (xvBP[0] + xvBP[3]) + 3*(xvBP[1] + xvBP[2])
                     + (dCoefficient1 * yvBP[0]) + (dCoefficient2  * yvBP[1])
                     + (dCoefficient3 * yvBP[2]);

  uint16_t retVal = (uint16_t)yvBP[3];
  
  if(retVal > 0x7FFF) //Clip those crazy negative values (or 65535 etc) to zero 
  {
    retVal = 0x0000;
  }
  
  return retVal;
}


uint16_t LMP90080::getTemp(uint16_t value) 
{
/* 
 *  Searches the 3850ppm/K RTD precalculated look-up table to find the temperature
 *  
 *  The RTD output code = (RTD resistance * Gain of 8 * 32768) / 1000 Ohm ref. resistor
 *  
 *  FOR loop implementation - benchmark shows much quicker than a binary search algorithm
 *  and marginally quicker than a WHILE loop
 *  
 */
for(int i = 0; i <= 450; i++) { //Step through the LUT
    
  if(LUT_3850[i] > value) {

  return i;
  }
 }
}


///////////////////////////////Private Functions///////////////////////////////

boolean LMP90080::writeURA(uint8_t address)
{
  /*Remember the URA is the upper nibble of the address byte
  * e.g 0x13 -> 0x1 is the URA, 0x03 is the LRA
  *
  * This function will be called with something like writeURA(0xYY)
  * and should be written to handle being called in the following 3 ways:
  * 1. writeURA(0x1) - the user has masked out the lower nibble themselves
  * 2. writeURA(0x13) - the user has send the full address
  * 3. writeURA(0x01) - the user has send the upper nibble shifted downwards = case 1
  */
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  SPI.transfer(0x10); //Write URA command

  //Did the user send the full address in the byte, just the upper nibble or the lower nibble?
  //Check if the upper nibble contains information (case 2 above)
  if(address > 0x0F)
  {
    SPI.transfer(0x00 | (address >> 4));
  }
  else //They must have only sent the upper nibble
  {
    SPI.transfer(address);
  }
  
  digitalWrite(_CS,HIGH);

  return true;
}


boolean LMP90080::writeLRA(uint8_t address, uint8_t data0)
{
  /* Page 33 of the LMP90080 Datasheet
   *  
   *  LRA = address to begin writing to
   *  
   *  'bytes' represents the size selection (SZ) for how many bytes are to be written
   *  0x0 = 1 byte
   *  0x1 = 2 bytes
   *  0x2 = 3 bytes
   *  0x3 = streaming - 3 or more bytes until !CS is taken high again
   *  
   *  With this first implementation, we're only writing one byte at a time ie SZ = 0x0
   */

  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  //Doesn't matter if the user send the full address or just the LRA as a nibble e.g. 0x02
  SPI.transfer(0x00 | (address & 0x0F));
  SPI.transfer(data0);
  digitalWrite(_CS,HIGH);

  return true;
}


boolean LMP90080::writeLRA(uint8_t address, uint8_t data0, uint8_t data1)
{
  /* Send two bytes of data*/
  
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  SPI.transfer(0x20 | (address & 0x0F));
  SPI.transfer(data0);
  SPI.transfer(data1);
  digitalWrite(_CS,HIGH);

  return true;
}


boolean LMP90080::writeLRA(uint8_t address, uint8_t data0, uint8_t data1, uint8_t data2)
{
  /* Send three bytes of data*/
  
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  SPI.transfer(0x40 | (address & 0x0F));
  SPI.transfer(data0);
  SPI.transfer(data1);
  SPI.transfer(data2);
  digitalWrite(_CS,HIGH);

  return true;
}



boolean LMP90080::readURA(uint8_t address)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  SPI.transfer(0x90); //Read URA command

  //Did the user send the full address in the byte, just the upper nibble or the lower nibble?
  //Check if the upper nibble contains information (case 2 above)
  if(address > 0x0F)
  {
    SPI.transfer(0x00 | (address >> 4));
  }
  else //They must have only sent the upper nibble
  {
    SPI.transfer(address);
  }
  
  digitalWrite(_CS,HIGH);

  return true;
}


uint8_t LMP90080::readLRA(uint8_t address)
{
  SPI.beginTransaction(SPISettings(CHIP_MHZ, MSBFIRST, SPI_MODE0)); 
    
  digitalWrite(_CS,LOW);
  //Doesn't matter if the user sends the full address or just the LRA as the lower nibble
  SPI.transfer(0x80 | (address & 0x0F)); //Read command
  uint8_t data = SPI.transfer(0x00);
  digitalWrite(_CS,HIGH);

  return data;
}
